---@type Tinkr
local Tinkr,
---@class Bastion
Bastion = ...

---@alias Bastion.APL.Action.Type "spell" | "item" | "apl" | "sequence" | "variable" | "function"
---@alias Bastion.APL.Action.Object Bastion.Spell | Bastion.Item | Bastion.APL

-- APL Action class
---@class Bastion.APL.Action
local APLAction = {}
APLAction.__index = APLAction


---@class Bastion.APL.Action.New.Params
---@field parent Bastion.APL
---@field type Bastion.APL.Action.Type
---@field object Bastion.APL.Action.Object
---@field condition? false | string | fun(self: Bastion.APL.Action, object: Bastion.APL.Action.Object, target?: Bastion.Unit): boolean
---@field onActionFunc? false | fun(object: Bastion.APL.Action.Object, target?: Bastion.Unit): any
---@field actionUsableFunc? false | fun(object: Bastion.APL.Action.Object, target?: Bastion.Unit): boolean
---@field executor? fun(self: Bastion.APL.Action): boolean
---@field target? Bastion.Unit | false
---@field targetIf? false | fun(object: Bastion.APL.Action.Object): Bastion.Unit

---@param params Bastion.APL.Action.New.Params
function APLAction:New(params)
    ---@class Bastion.APL.Action
    local self = setmetatable({}, APLAction)
    self.parent = params.parent
    self.index = #self.parent.actions + 1
    self.type = params.type
    self.object = params.object
    self.condition = params.condition or false
    self.onActionFunc = params.onActionFunc or false
    self.actionUsableFunc = params.actionUsableFunc or false
    self.executor = params.executor or false
    self.target = params.target or false
    self.targetIf = params.targetIf or false
    return self
end

function APLAction:Execute()
    return self.executor and self.executor(self) or false
end

---@class Bastion.APL.Action.New.Spell.Params
---@field parent Bastion.APL
---@field spell Bastion.Spell
---@field condition? false | string | fun(self: Bastion.APL.Action, object: Bastion.Spell, target?: Bastion.Unit): boolean
---@field target? Bastion.Unit
---@field targetIf? fun(self: Bastion.Spell): Bastion.Unit

---@param params Bastion.APL.Action.New.Spell.Params
function APLAction:NewSpell(params)
    return self:New({
        parent = params.parent,
        type = "spell",
        object = params.spell,
        condition = params.condition or false,
        target = params.target or params.spell:GetTarget(),
        targetIf = params.targetIf or false,
        onActionFunc = params.spell.OnCastFunc or false,
        actionUsableFunc = params.spell.CastableIfFunc or false,
        executor = function(this)
            local target = this.targetIf and this.targetIf(this.object) or this.target or nil
            return this.object:CastableIf(this.actionUsableFunc):OnCast(this.onActionFunc):Cast(target, this.condition)
        end,
    })
end

---@class Bastion.APL.Action.New.Item.Params
---@field parent Bastion.APL
---@field item Bastion.Item
---@field condition? false | string | fun(self: Bastion.APL.Action, object: Bastion.Item, target?: Bastion.Unit): boolean
---@field target? Bastion.Unit
---@field targetIf? fun(self: Bastion.Item): Bastion.Unit

---@param params Bastion.APL.Action.New.Item.Params
function APLAction:NewItem(params)
    return self:New({
        parent = params.parent,
        type = "item",
        object = params.item,
        condition = params.condition or false,
        target = params.target or params.item:GetTarget() or false,
        targetIf = params.targetIf or false,
        onActionFunc = params.item.OnUseFunc or false,
        actionUsableFunc = params.item.UsableIfFunc or false,
        executor = function(this)
            return this.object:UsableIf(this.actionUsableFunc):OnUse(this.onActionFunc):Use(
                this.targetIf and this.targetIf(this.object) or this.target or nil, this.condition)
        end,
    })
end

---@class Bastion.APL.Action.New.APL.Params
---@field parent Bastion.APL
---@field apl Bastion.APL
---@field condition? false | string | fun(self: Bastion.APL.Action, object: Bastion.APL): boolean

function APLAction:NewAPL(params)
    return self:New({
        parent = params.parent,
        type = "apl",
        object = params.apl,
        condition = params.condition or false,
        executor = function(this)
            return this.object:Execute()
        end,
    })
end

Bastion.APLAction = APLAction
